"""Commonly used colors."""

from typing import Final

from pygments.style import Style
from pygments.token import (
    Comment,
    Error,
    Generic,
    Keyword,
    Literal,
    Name,
    Number,
    Operator,
    String,
    Text,
    Token,
)

from .color_types import AColor, Color, ColorTheme

__all__ = [
    "ABLACK",
    "ABLUE",
    "ACYAN",
    "AGREEN",
    "AMAGENTA",
    "ARED",
    "AWHITE",
    "AYELLOW",
    "BLACK",
    "BLUE",
    "CYAN",
    "NEPTUNE_PRIMARY_BG",
    "NEPTUNE_PRIMARY_FG",
    "GREEN",
    "MAGENTA",
    "NEPTUNE_THEME",
    "RED",
    "TRANSPARENT",
    "WHITE",
    "YELLOW",
    "Neptune",
]

WHITE: Final = Color(255, 255, 255)
"""White."""

BLACK: Final = Color(0, 0, 0)
"""Black."""

RED: Final = Color(255, 0, 0)
"""Red."""

GREEN: Final = Color(0, 255, 0)
"""Green."""

BLUE: Final = Color(0, 0, 255)
"""Blue."""

YELLOW: Final = Color(255, 255, 0)
"""Yellow."""

CYAN: Final = Color(0, 255, 255)
"""Cyan."""

MAGENTA: Final = Color(255, 0, 255)
"""Magenta."""

AWHITE: Final = AColor(255, 255, 255)
"""Opaque white."""

ABLACK: Final = AColor(0, 0, 0)
"""Opaque black."""

ARED: Final = AColor(255, 0, 0)
"""Opaque red."""

AGREEN: Final = AColor(0, 255, 0)
"""Opaque green."""

ABLUE: Final = AColor(0, 0, 255)
"""Opaque blue."""

AYELLOW: Final = AColor(255, 255, 0)
"""Opaque yellow."""

ACYAN: Final = AColor(0, 255, 255)
"""Opaque cyan."""

AMAGENTA: Final = AColor(255, 0, 255)
"""Opaque magenta."""

TRANSPARENT: Final = AColor(0, 0, 0, 0)
"""Transparent black."""

NEPTUNE_THEME: Final[ColorTheme] = {
    "primary_fg": "f6a7a9",
    "primary_bg": "070c25",
    "text_pad_line_highlight_fg": "f6a7a9",
    "text_pad_line_highlight_bg": "0c0e30",
    "text_pad_selection_highlight_fg": "f6a7a9",
    "text_pad_selection_highlight_bg": "0f1847",
    "textbox_primary_fg": "fff0f6",
    "textbox_primary_bg": "070c25",
    "textbox_selection_highlight_fg": "fff0f6",
    "textbox_selection_highlight_bg": "0f1847",
    "textbox_placeholder_fg": "2a3a92",
    "textbox_placeholder_bg": "070c25",
    "button_normal_fg": "dde4ed",
    "button_normal_bg": "2a3ca0",
    "button_hover_fg": "fff0f6",
    "button_hover_bg": "3248c0",
    "button_press_fg": "fff0f6",
    "button_press_bg": "c4a219",
    "button_disallowed_fg": "272b40",
    "button_disallowed_bg": "070c25",
    "menu_item_hover_fg": "f2babc",
    "menu_item_hover_bg": "111834",
    "menu_item_selected_fg": "ecf3ff",
    "menu_item_selected_bg": "1b244b",
    "menu_item_disallowed_fg": "272b40",
    "menu_item_disallowed_bg": "070c25",
    "titlebar_normal_fg": "ffe0df",
    "titlebar_normal_bg": "070c25",
    "titlebar_inactive_fg": "7d6b71",
    "titlebar_inactive_bg": "070c25",
    "data_table_sort_indicator_fg": "ecf3ff",
    "data_table_sort_indicator_bg": "070c25",
    "data_table_hover_fg": "f6a7a9",
    "data_table_hover_bg": "111834",
    "data_table_stripe_fg": "f6a7a9",
    "data_table_stripe_bg": "0b1238",
    "data_table_stripe_hover_fg": "f6a7a9",
    "data_table_stripe_hover_bg": "0f184a",
    "data_table_selected_fg": "ecf3ff",
    "data_table_selected_bg": "111f5e",
    "data_table_selected_hover_fg": "ecf3ff",
    "data_table_selected_hover_bg": "1b244b",
    "progress_bar_fg": "ffe0df",
    "progress_bar_bg": "2a3ca0",
    "markdown_link_fg": "376cff",
    "markdown_link_bg": "070c25",
    "markdown_link_hover_fg": "4668ff",
    "markdown_link_hover_bg": "070c25",
    "markdown_inline_code_fg": "806ae5",
    "markdown_inline_code_bg": "080b1a",
    "markdown_quote_fg": "2054e2",
    "markdown_quote_bg": "0c1b4b",
    "markdown_title_fg": "cfd1d4",
    "markdown_title_bg": "292a2d",
    "markdown_image_fg": "f6a7a9",
    "markdown_image_bg": "0c1540",
    "markdown_block_code_bg": "080b1a",
    "markdown_quote_block_code_bg": "11265d",
    "markdown_header_bg": "030612",
    "scroll_view_scrollbar": "070c25",
    "scroll_view_indicator_normal": "0e1843",
    "scroll_view_indicator_hover": "111e4f",
    "scroll_view_indicator_press": "172868",
    "window_border_normal": "122162",
    "window_border_inactive": "282c3e",
}
"""Neptune color theme."""

NEPTUNE_PRIMARY_FG: Final = Color.from_hex(NEPTUNE_THEME["primary_fg"])
"""The primary foreground color for the Neptune color theme."""

NEPTUNE_PRIMARY_BG: Final = Color.from_hex(NEPTUNE_THEME["primary_bg"])
"""The primary background color for the Neptune color theme."""


class Neptune(Style):
    """
    Default syntax highlight color theme.

    This style is adapted from
    https://github.com/yl92/paddy-color-theme/blob/master/themes/upright/Paddy-neptune-upright-color-theme.json.
    """

    name = "neptune"

    background_color = "#070c25"
    highlight_color = "#5f32db"

    line_number_special_color = "#2f3adf"
    line_number_special_background_color = "#5f32db"

    line_number_color = "#527bff"
    line_number_background_color = "#070c25"

    styles = {
        Token: "#f6a7a9",
        Error: "bold #d32e11",
        Keyword: "#5bb177",
        Keyword.Constant: "bold #91a5bd",
        Keyword.Pseudo: "#e28f37",
        Name: "#f6a7a9",
        Name.Class: "bold #de9138",
        Name.Constant: "#17c78c",
        Name.Decorator: "#d85757",
        Name.Exception: "bold #de9138",
        Name.Function: "#d85757",
        Name.Label: "#f6a7a9",
        Name.Property: "#f6a7a9",
        Name.Tag: "bold #309388",
        Name.Variable: "#f6a7a9",
        Literal: "#da7763",
        String: "#da7763",
        String.Affix: "#c75974",
        String.Escape: "#62A594",
        String.Regex: "#d46fbe",
        Number: "#e47559",
        Comment: "#554492",
        Comment.Special: "#978d7a",
        Operator: "#5bb177",
        Generic: "#f6a7a9",
        Generic.Deleted: "bold bg:#cf5076 #309388",
        Generic.Emph: "italic",
        Generic.Error: "bold #d32e11",
        Generic.Heading: "#da5656",
        Generic.Inserted: "bg:#4d67ff #da7763",
        Generic.Output: "#69a097",
        Generic.Strong: "bold",
        Generic.EmphStrong: "bold italic",
        Generic.Subheading: "#da5656",
        Generic.Traceback: "bold #de9138",
        Generic.Underline: "underline",
        Text: "#f6a7a9",
    }



"""Build cython extensions."""

import numpy as np
from Cython.Build import cythonize
from setuptools import setup

setup(
    ext_modules=cythonize(
        [
            "src/batgrl/_rendering.pyx",
            "src/batgrl/_sixel.pyx",
            "src/batgrl/gadgets/_raycasting.pyx",
            "src/batgrl/gadgets/_shadow_casting.pyx",
            "src/batgrl/geometry/regions.pyx",
            "src/batgrl/terminal/_fbuf.pyx",
            "src/batgrl/terminal/vt100_terminal.pyx",
        ]
    ),
    include_dirs=[np.get_include()],
)



"""
Logging related classes and functions.

Because default logging opens stdout/stderr which disables batgrl output, logging must
be setup with a file handler instead. To ensure a file handler has been added, all
library logging should use ``get_logger`` (essentially, just an alias of stdlib
``logging.getLogger``) from this module.

Additional log levels, ``ANSI`` and ``EVENTS``, are added to standard logging levels.
"""
# Inspiration taken from:
# <https://github.com/python-discord/bot-core/blob/main/pydis_core/utils/logging.py>.
# python-discord/bot-core is licensed under the MIT License.

import logging
import typing
from enum import IntEnum
from pathlib import Path
from typing import Final

__all__ = ["LogLevel", "get_logger"]

LOG_FILE = Path() / "batgrl.log"  # TODO: Customize path.
ANSI_LEVEL: Final = 3
"""
``ANSI`` log level.

Show generated ANSI in logs.
"""
EVENTS_LEVEL: Final = 5
"""
``EVENTS`` log level.

Show generated input events.
"""


class LogLevel(IntEnum):
    """Standard logging levels with additional ``ANSI`` and ``EVENTS`` levels."""

    ANSI = ANSI_LEVEL
    """Show generated ANSI in logs."""
    EVENTS = EVENTS_LEVEL
    """Show generated input events."""
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


if typing.TYPE_CHECKING:
    LoggerClass = logging.Logger
else:
    LoggerClass = logging.getLoggerClass()


class CustomLogger(LoggerClass):
    """A standard logger with methods for ``ANSI`` and ``EVENTS`` logging."""

    def ansi(self, msg: str, *args, **kwargs) -> None:
        """Log the given message with the severity ``"ANSI"``."""
        if self.isEnabledFor(logging.ANSI):
            self.log(logging.ANSI, msg, *args, **kwargs)

    def events(self, msg: str, *args, **kwargs) -> None:
        """Log the given message with the severity ``"EVENTS"``."""
        if self.isEnabledFor(logging.EVENTS):
            self.log(logging.EVENTS, msg, *args, **kwargs)

    def is_enabled_for(self, log_level: LogLevel | str) -> bool:
        """Similar to ``Logger.isEnabledFor``, but also accepts level names."""
        if isinstance(log_level, str):
            return self.isEnabledFor(logging._nameToLevel[log_level])
        return self.isEnabledFor(log_level)


def get_logger(name: str | None = None) -> CustomLogger:
    """
    Return a logger with a specified name.

    Because there are several entry points into batgrl and default logging opens
    stderr/stdout which will disable batgrl output, all library logging must use this
    alias of ``logging.getLogger`` to ensure logging has been set to output to a file.

    Parameters
    ----------
    name : str | None, default: None
        Name of logger. If not specified, return the root logger.

    Returns
    -------
    CustomLogger
        The specified logger.
    """
    return typing.cast(CustomLogger, logging.getLogger(name))


logging.ANSI: Final = ANSI_LEVEL
logging.EVENTS: Final = EVENTS_LEVEL
logging.addLevelName(ANSI_LEVEL, "ANSI")
logging.addLevelName(EVENTS_LEVEL, "EVENTS")
logging.setLoggerClass(CustomLogger)

log_format = logging.Formatter(
    "{asctime} | {levelname} | {name} | {message}", style="{"
)

log_handler = logging.FileHandler(LOG_FILE, mode="w", encoding="utf-8")
log_handler.setFormatter(log_format)

logger = logging.getLogger("batgrl")
logger.addHandler(log_handler)
logger.setLevel(LogLevel.INFO)



"""Button behavior for a gadget."""

from typing import Literal

__all__ = ["ButtonState", "ButtonBehavior"]

ButtonState = Literal["normal", "hover", "down", "disallowed"]
"""Button behavior states."""


class ButtonBehavior:
    """
    Button behavior for a gadget.

    A button has four states: "normal", "hover", "down", and "disallowed".

    When a button's state changes one of the following methods are called:
    - :meth:`update_normal`
    - :meth:`update_hover`
    - :meth:`update_down`
    - :meth:`update_disallowed`

    When a button is released, the :meth:`on_release` method is called.

    Parameters
    ----------
    always_release : bool, default: False
        Whether a mouse up event outside the button will trigger it.

    Attributes
    ----------
    always_release : bool
        Whether a mouse up event outside the button will trigger it.
    button_state : ButtonState
        Current button state.

    Methods
    -------
    on_release()
        Triggered when a button is released.
    update_normal()
        Paint the normal state.
    update_hover()
        Paint the hover state.
    update_down()
        Paint the down state.
    update_disallowed()
        Paint the disallowed state.
    """

    def __init__(self, *, always_release: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.always_release = always_release
        self.button_state: ButtonState = "normal"

    @property
    def button_state(self) -> ButtonState:
        """Current button state."""
        return self._button_state

    @button_state.setter
    def button_state(self, button_state: ButtonState):
        dispatch = {
            "normal": self.update_normal,
            "hover": self.update_hover,
            "down": self.update_down,
            "disallowed": self.update_disallowed,
        }
        if button_state not in dispatch:
            button_state = "normal"

        self._button_state = button_state
        if self.root:
            dispatch[button_state]()

    def on_add(self):
        """Paint normal state on add."""
        super().on_add()
        self.update_normal()

    def on_mouse(self, mouse_event) -> bool | None:
        """Determine button state from mouse event."""
        if self.button_state == "disallowed":
            return False

        if super().on_mouse(mouse_event):
            return True

        collides = self.collides_point(mouse_event.pos)

        if mouse_event.event_type == "mouse_down":
            if collides:
                self.button_state = "down"
                return True

        elif mouse_event.event_type == "mouse_up" and self.button_state == "down":
            if collides:
                self.on_release()
                self.button_state = "hover"
                return True

            self.button_state = "normal"

            if self.always_release:
                self.on_release()
                return True

        if not collides and self.button_state == "hover":
            self.button_state = "normal"
        elif collides and self.button_state == "normal":
            self.button_state = "hover"

    def on_release(self):
        """Triggered when button is released."""

    def update_normal(self):
        """Paint the normal state."""

    def update_hover(self):
        """Paint the hover state."""

    def update_down(self):
        """Paint the down state."""

    def update_disallowed(self):
        """Paint the disallowed state."""



"""Base class for all gadgets."""

from __future__ import annotations

import asyncio
from collections.abc import Callable, Iterator, MutableSequence, Sequence
from functools import wraps
from itertools import count
from numbers import Real
from time import perf_counter
from types import MappingProxyType
from typing import TYPE_CHECKING, Final, Literal, TypedDict
from weakref import WeakKeyDictionary

import numpy as np
from numpy.typing import NDArray

from ..geometry import EASINGS, Easing, Point, Region, Size, clamp, lerp, round_down
from ..terminal.events import FocusEvent, KeyEvent, MouseEvent, PasteEvent
from ..text_tools import Cell, new_cell

if TYPE_CHECKING:
    from ._root import _Root

__all__ = [
    "Anchor",
    "Cell",
    "Easing",
    "Point",
    "PosHint",
    "Region",
    "Size",
    "SizeHint",
    "Gadget",
    "bindable",
    "new_cell",
    "clamp",
    "lerp",
]

_UID: Final = count(1)

Anchor = Literal[
    "top-left",
    "top",
    "top-right",
    "left",
    "center",
    "right",
    "bottom-left",
    "bottom",
    "bottom-right",
]
"""Point of gadget attached to a pos hint."""


class _GadgetList(MutableSequence):
    """
    A sequence of sibling gadgets.

    Gadget regions are invalidated when ``_GadgetList`` is mutated.
    """

    def __init__(self) -> None:
        self._gadgets: list[Gadget] = []

    def __len__(self) -> int:
        return len(self._gadgets)

    def __getitem__(self, index: int) -> Gadget:
        return self._gadgets[index]

    def __setitem__(self, *_) -> None:
        raise NotImplementedError("_GadgetList.__setitem__ not implemented.")

    def __delitem__(self, index: int) -> None:
        self._gadgets[index]._invalidate_regions()
        del self._gadgets[index]

    def insert(self, index: int, gadget: Gadget) -> None:
        gadget._invalidate_regions()
        self._gadgets.insert(index, gadget)

    def __iter__(self) -> Iterator[Gadget]:
        return iter(self._gadgets)

    def copy(self) -> list[Gadget]:
        """
        Return a copy of the internal list of gadgets.

        Note this doesn't return a ``_GadgetList``.
        """
        return self._gadgets.copy()


class PosHint(TypedDict, total=False):
    """
    A position hint.

    A pos hint allows a gadget to automatically position itself when added to the
    gadget tree or when its parent resizes. ``y_hint`` controls vertical position and
    ``x_hint`` horizontal. ``anchor`` determines which point of the gadget is attached
    to the pos hint. For instance, in the diagram below, if the ``y_hint`` and
    ``x_hint`` are both ``0.5`` the pos hint will be at point ``c`` on the parent
    (50% of the parent's height and width). Additionally, if the anchor is ``"top"``,
    the anchor will be at point ``a`` on the gadget::

             parent
        +---------------+
        |               |    +---a---+
        |       c       |    |gadget |
        |               |    +-------+
        +---------------+


    Subsequently, ``a`` will be aligned with ``c``, so that the gadget is positioned as
    below::

             parent
        +---------------+
        |               |
        |   +-------+   |
        |   |gadget |   |
        +---+-------+---+

    The additional parameters ``y_offset`` and ``x_offset`` allow one to translate the
    gadget by some integer offsets after the pos hint has been applied.

    Attributes
    ----------
    anchor : Anchor | tuple[float, float]
        Determines which point is attached to the pos hint.
    y_hint : float | None
        Vertical position as a proportion of parent's height.
    x_hint : float | None
        Horizontal position as a proportion of parent's width.
    y_offset : int
        Vertical offset after pos hint is applied.
    x_offset : int
        Horizontal offset after pos hint is applied.
    """

    anchor: Anchor | tuple[float, float]
    """Determines which point is attached to the pos hint."""
    y_hint: float | None
    """Vertical position as a proportion of parent's height."""
    x_hint: float | None
    """Horizontal position as a proportion of parent's width."""
    y_offset: int
    """Vertical offset after pos hint is applied."""
    x_offset: int
    """Horizontal offset after pos hint is applied."""


class SizeHint(TypedDict, total=False):
    """
    A size hint.

    A size hint allows a gadget to automatically size itself when added to the gadget
    tree or when its parent resizes. ``height_hint`` is the proportion of the parent's
    height the gadget's height will be and ``width_hint`` the proportion of the parent's
    width. Additional parameters ``height_offset`` and ``width_offset`` allow adjusting
    the size by some integer amount after the size hint has been applied. If given,
    ``min_height``, ``max_height``, ``min_width``, and ``max_width`` will prevent the
    gadget from sizing too small or too large.

    Attributes
    ----------
    height_hint : float | None
        Height as a proportion of parent's height.
    width_hint : float | None
        Width as a proportion of parent's width.
    height_offset : int
        Height offset after height-hint is applied.
    width_offset : int
        Width offset after width-hint is applied.
    max_height : int | None
        Maximum allowed height.
    min_height : int | None
        Minimum allowed height.
    max_width : int | None
        Maximum allowed width.
    min_width : int | None
        Minimum allowed width.
    """

    height_hint: float | None
    """Height as a proportion of parent's height."""
    width_hint: float | None
    """Width as a proportion of parent's width."""
    height_offset: int
    """Height offset after height-hint is applied."""
    width_offset: int
    """Width offset after width-hint is applied."""
    max_height: int | None
    """Maximum allowed height."""
    min_height: int | None
    """Minimum allowed height."""
    max_width: int | None
    """Maximum allowed width."""
    min_width: int | None
    """Minimum allowed width."""


_ANCHOR_TO_COORD: Final[dict[Anchor, tuple[float, float]]] = {
    "top-left": (0.0, 0.0),
    "top": (0.0, 0.5),
    "top-right": (0.0, 1.0),
    "left": (0.5, 0.0),
    "center": (0.5, 0.5),
    "right": (0.5, 1.0),
    "bottom-left": (1.0, 0.0),
    "bottom": (1.0, 0.5),
    "bottom-right": (1.0, 1.0),
}
"""Anchor to coordinate."""
_DEFAULT_POS_HINT: Final[PosHint] = {
    "anchor": "center",
    "y_hint": None,
    "x_hint": None,
    "x_offset": 0,
    "y_offset": 0,
}
"""The default pos hint."""
_DEFAULT_SIZE_HINT: Final[SizeHint] = {
    "height_hint": None,
    "width_hint": None,
    "height_offset": 0,
    "width_offset": 0,
    "max_height": None,
    "min_height": None,
    "max_width": None,
    "min_width": None,
}
"""The default size hint."""


def _normalize_pos_hint(pos_hint: PosHint) -> PosHint:
    normal_hint = _DEFAULT_POS_HINT | pos_hint
    if normal_hint["y_hint"] is not None:
        normal_hint["y_hint"] = float(normal_hint["y_hint"])
    if normal_hint["x_hint"] is not None:
        normal_hint["x_hint"] = float(normal_hint["x_hint"])
    if isinstance(normal_hint["anchor"], str):
        normal_hint["anchor"] = _ANCHOR_TO_COORD[normal_hint["anchor"]]
    return normal_hint


def _normalize_size_hint(size_hint: SizeHint) -> SizeHint:
    normal_hint = _DEFAULT_SIZE_HINT | size_hint
    if normal_hint["height_hint"] is not None:
        normal_hint["height_hint"] = float(normal_hint["height_hint"])
    if normal_hint["width_hint"] is not None:
        normal_hint["width_hint"] = float(normal_hint["width_hint"])
    return normal_hint


def bindable(setter):
    """Decorate property setters to make them bindable."""
    instances: WeakKeyDictionary[Gadget, dict[int, Callable[[], None]]]
    instances = WeakKeyDictionary()

    @wraps(setter)
    def wrapper(self, *args, **kwargs):
        setter(self, *args, **kwargs)
        if bindings := instances.get(self):
            for callback in bindings.values():
                callback()

    wrapper.instances = instances

    return wrapper


class Gadget:
    r"""
    Base class for creating gadgets.

    Parameters
    ----------
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent : Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    __bindings: dict[int, str] = {}
    """UID to property name mapping."""

    def __init__(
        self,
        *,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.parent: Gadget | None = None
        """The gadget's parent."""
        self.children: _GadgetList = _GadgetList()
        """The gadget's children."""

        h, w = size
        self._size = Size(clamp(h, 0, None), clamp(w, 0, None))
        """Size of gadget."""
        self._pos = Point(*pos)
        """Position relative to parent."""
        self._size_hint: SizeHint = _normalize_size_hint(size_hint or {})
        """Gadget's size as a proportion of its parent's size."""
        self._pos_hint: PosHint = _normalize_pos_hint(pos_hint or {})
        """Gadget's position as a proportion of its parent's size."""
        self._is_transparent = is_transparent
        """Whether gadget is transparent."""
        self._is_visible = is_visible
        """Whether gadget is visible."""
        self._is_enabled = is_enabled
        """Whether gadget is enabled."""
        self._region: Region = Region()
        """The visible portion of the gadget on the screen."""

    def __repr__(self):
        return (
            f"{type(self).__name__}(size={self.size}, pos={self.pos}, "
            f"is_transparent={self.is_transparent}, is_visible={self.is_visible}, "
            f"is_enabled={self.is_enabled})"
        )

    @property
    def size(self) -> Size:
        """Size of gadget."""
        return self._size

    @size.setter
    @bindable
    def size(self, size: Size):
        if size == self._size:
            self._apply_pos_hints()
            return

        h, w = size
        size = Size(clamp(int(h), 0, None), clamp(int(w), 0, None))
        if self.root is None:
            self._size = size
        else:
            with self.root._render_lock:
                self._size = size
                self._invalidate_regions()

        self._apply_pos_hints()
        for child in self.children:
            child.apply_hints()
        self.on_size()

    @property
    def height(self) -> int:
        """Height of gadget."""
        return self._size[0]

    @height.setter
    def height(self, height: int):
        self.size = height, self.width

    rows = height
    """Alias for :attr:`height`."""

    @property
    def width(self) -> int:
        """Width of gadget."""
        return self._size[1]

    @width.setter
    def width(self, width: int):
        self.size = self.height, width

    columns = width
    """Alias for :attr:`width`."""

    @property
    def pos(self) -> Point:
        """Position relative to parent."""
        return self._pos

    @pos.setter
    @bindable
    def pos(self, pos: Point):
        y, x = pos
        pos = Point(int(y), int(x))

        if self.root is None:
            self._pos = pos
        else:
            with self.root._render_lock:
                self._pos = pos
                self._invalidate_regions()

    @property
    def top(self) -> int:
        """y-coordinate of top of gadget."""
        return self._pos[0]

    @top.setter
    def top(self, top: int):
        self.pos = top, self.left

    y = top
    """y-coordinate of top of gadget."""

    @property
    def left(self) -> int:
        """x-coordinate of left side of gadget."""
        return self._pos[1]

    @left.setter
    def left(self, left: int):
        self.pos = self.top, left

    x = left
    """x-coordinate of left side of gadget."""

    @property
    def bottom(self) -> int:
        """y-coordinate of bottom of gadget."""
        return self.top + self.height

    @bottom.setter
    def bottom(self, bottom: int):
        self.top = bottom - self.height

    @property
    def right(self) -> int:
        """x-coordinate of right side of gadget."""
        return self.left + self.width

    @right.setter
    def right(self, right: int):
        self.left = right - self.width

    @property
    def center(self) -> Point:
        """Position of center of gadget."""
        return self.pos + self.size.center

    @center.setter
    def center(self, center: Point):
        cy, cx = center
        h, w = self.size
        self.pos = Point(cy - h // 2, cx - w // 2)

    @property
    def absolute_pos(self) -> Point:
        """Absolute position on screen."""
        if self.parent is None:
            return self.pos
        return self.pos + self.parent.absolute_pos

    @property
    def size_hint(self) -> SizeHint:
        """Gadget's size as a proportion of its parent's size."""
        return MappingProxyType(self._size_hint)

    @size_hint.setter
    def size_hint(self, size_hint: SizeHint):
        self._size_hint = _normalize_size_hint(size_hint)
        self.apply_hints()

    @property
    def pos_hint(self) -> PosHint:
        """Gadget's position as a proportion of its parent's size."""
        return MappingProxyType(self._pos_hint)

    @pos_hint.setter
    def pos_hint(self, pos_hint: PosHint):
        self._pos_hint = _normalize_pos_hint(pos_hint)
        self.apply_hints()

    @property
    def is_transparent(self) -> bool:
        """Whether gadget is transparent."""
        return self._is_transparent

    @is_transparent.setter
    @bindable
    def is_transparent(self, is_transparent: bool):
        if is_transparent != self._is_transparent:
            self._is_transparent = is_transparent
            if self.root is not None:
                self._invalidate_regions()
            self.on_transparency()

    @property
    def is_visible(self) -> bool:
        """
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
        """
        return self._is_visible

    @is_visible.setter
    @bindable
    def is_visible(self, is_visible: bool):
        if is_visible != self._is_visible:
            self._is_visible = is_visible
            if self.root is not None:
                self._invalidate_regions()

    @property
    def is_enabled(self) -> bool:
        """
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.
        """
        return self._is_enabled

    @is_enabled.setter
    @bindable
    def is_enabled(self, is_enabled: bool):
        if is_enabled != self._is_enabled:
            self._is_enabled = is_enabled
            if self.root is not None:
                self._invalidate_regions()

    @property
    def root(self) -> _Root | None:
        """Return the root gadget if connected to gadget tree."""
        return self.parent and self.parent.root

    @property
    def app(self):
        """The running app."""
        return self.root.app

    def _render(
        self, cells: NDArray[Cell], graphics: NDArray[np.uint8], kind: NDArray[np.uint8]
    ) -> None:
        """Render visible region of gadget."""

    def dispatch_key(self, key_event: KeyEvent) -> bool | None:
        """
        Dispatch a key press event until handled.

        A key press event is handled if a handler returns ``True``.

        Parameters
        ----------
        key_event : KeyEvent
            The key event to dispatch.

        Returns
        -------
        bool | None
            Whether the dispatch was handled.
        """
        return any(
            gadget.dispatch_key(key_event)
            for gadget in reversed(self.children)
            if gadget.is_enabled
        ) or self.on_key(key_event)

    def dispatch_mouse(self, mouse_event: MouseEvent) -> bool | None:
        """
        Dispatch a mouse event until handled.

        A mouse event is handled if a handler returns ``True``.

        Parameters
        ----------
        mouse_event : MouseEvent
            The mouse event to dispatch.

        Returns
        -------
        bool | None
            Whether the dispatch was handled.
        """
        return any(
            gadget.dispatch_mouse(mouse_event)
            for gadget in reversed(self.children)
            if gadget.is_enabled
        ) or self.on_mouse(mouse_event)

    def dispatch_paste(self, paste_event: PasteEvent) -> bool | None:
        """
        Dispatch a paste event until handled.

        A paste event is handled if a handler returns ``True``.

        Parameters
        ----------
        paste_event : PasteEvent
            The paste event to dispatch.

        Returns
        -------
        bool | None
            Whether the dispatch was handled.
        """
        return any(
            gadget.dispatch_paste(paste_event)
            for gadget in reversed(self.children)
            if gadget.is_enabled
        ) or self.on_paste(paste_event)

    def dispatch_terminal_focus(self, focus_event: FocusEvent) -> bool | None:
        """
        Dispatch a focus event until handled.

        A focus event is handled if a handler returns ``True``.

        Parameters
        ----------
        focus_event : FocusEvent
            The focus event to dispatch.

        Returns
        -------
        bool | None
            Whether the dispatch was handled.
        """
        return any(
            gadget.dispatch_terminal_focus(focus_event)
            for gadget in reversed(self.children)
            if gadget.is_enabled
        ) or self.on_terminal_focus(focus_event)

    def _invalidate_regions(self) -> None:
        """Invalidate regions."""
        if self.root is not None:
            self.root._regions_valid = False

    def apply_hints(self) -> None:
        """
        Apply size and pos hints.

        This is called automatically when the gadget is added to the gadget tree and
        when the gadget's parent's size changes.
        """
        if self.parent is None:
            return

        parent_height, parent_width = self.parent.size
        height, width = self.size
        if self._size_hint["height_hint"] is not None:
            height = clamp(
                round_down(parent_height * self._size_hint["height_hint"])
                + self._size_hint["height_offset"],
                self._size_hint["min_height"],
                self._size_hint["max_height"],
            )

        if self._size_hint["width_hint"] is not None:
            width = clamp(
                round_down(parent_width * self._size_hint["width_hint"])
                + self._size_hint["width_offset"],
                self._size_hint["min_width"],
                self._size_hint["max_width"],
            )

        self.size = height, width  # `size` setter will call `_apply_pos_hints()`.

    def _apply_pos_hints(self) -> None:
        if self.parent is None:
            return

        parent_height, parent_width = self.parent.size
        height, width = self.size
        y, x = self.pos
        y_anchor, x_anchor = self._pos_hint["anchor"]

        if self._pos_hint["y_hint"] is not None:
            y = (
                round_down(parent_height * self._pos_hint["y_hint"])
                - round_down(height * y_anchor)
                + self._pos_hint["y_offset"]
            )

        if self._pos_hint["x_hint"] is not None:
            x = (
                round_down(parent_width * self._pos_hint["x_hint"])
                - round_down(width * x_anchor)
                + self._pos_hint["x_offset"]
            )
        self.pos = y, x

    def to_local(self, point: Point) -> Point:
        """
        Convert point in absolute coordinates to local coordinates.

        Parameters
        ----------
        point : Point
            Point in absolute (screen) coordinates.

        Returns
        -------
        Point
            The point in local coordinates.
        """
        return point - self.absolute_pos

    def collides_point(self, point: Point) -> bool:
        """
        Return true if point collides with visible portion of gadget.

        Parameters
        ----------
        point : Point
            A point.

        Returns
        -------
        bool
            Whether point collides with gadget.
        """
        if self.root is None or not (self.is_visible and self.is_enabled):
            return False

        return point in self._region or any(
            point in child._region
            for child in self.walk()
            if child.is_visible and child.is_enabled
        )

    def collides_gadget(self, other: Gadget) -> bool:
        """
        Return true if other is within gadget's bounding box.

        Parameters
        ----------
        other : Gadget
            Another gadget.

        Returns
        -------
        bool
            Whether other collides with gadget.
        """
        self_top, self_left = self.absolute_pos
        self_bottom = self_top + self.height
        self_right = self.left + self.width

        other_top, other_left = other.absolute_pos
        other_bottom = other_top + other.height
        other_right = other_left + other.width

        return not (
            self_top >= other_bottom
            or other_top >= self_bottom
            or self_left >= other_right
            or other_left >= self_right
        )

    def pull_to_front(self) -> None:
        """Move gadget to end of gadget stack so that it is drawn last."""
        if self.parent is not None:
            self.parent.children.remove(self)
            self.parent.children.append(self)

    def walk(self) -> Iterator[Gadget]:
        """
        Yield all descendents of this gadget (preorder traversal).

        Yields
        ------
        Gadget
            A descendent of this gadget.
        """
        for child in self.children:
            yield child
            yield from child.walk()

    def walk_reverse(self) -> Iterator[Gadget]:
        """
        Yield all descendents of this gadget (reverse postorder traversal).

        Yields
        ------
        Gadget
            A descendent of this gadget.
        """
        for child in reversed(self.children):
            yield from child.walk_reverse()
            yield child

    def ancestors(self) -> Iterator[Gadget]:
        """
        Yield all ancestors of this gadget.

        Yields
        ------
        Gadget
            An ancestor of this gadget.
        """
        if self.parent:
            yield self.parent
            yield from self.parent.ancestors()

    def add_gadget(self, gadget: Gadget) -> None:
        """
        Add a child gadget.

        Parameters
        ----------
        gadget : Gadget
            A gadget to add as a child.
        """
        self.children.append(gadget)
        gadget.parent = self
        self._invalidate_regions()

        if self.root is not None:
            gadget.on_add()

    def add_gadgets(self, *gadgets: Gadget) -> None:
        r"""
        Add multiple child gadgets.

        Parameters
        ----------
        \*gadgets : Gadget
            Gadgets to add as children. Can also accept a single iterable of gadgets.
        """
        if len(gadgets) == 1 and not isinstance(gadgets[0], Gadget):
            # Assume item is an iterable of gadgets.
            gadgets = gadgets[0]

        for gadget in gadgets:
            self.add_gadget(gadget)

    def remove_gadget(self, gadget: Gadget) -> None:
        """
        Remove a child gadget.

        Parameters
        ----------
        gadget : Gadget
            The gadget to remove from children.
        """
        if self.root is not None:
            gadget.on_remove()

        self.children.remove(gadget)
        gadget.parent = None
        self._invalidate_regions()

    def prolicide(self) -> None:
        """Recursively remove all children."""
        for child in self.children.copy():
            child.destroy()

    def destroy(self) -> None:
        """Remove this gadget and recursively remove all its children."""
        self.prolicide()
        if self.parent:
            self.parent.remove_gadget(self)

    def bind(self, prop: str, callback: Callable[[], None]) -> int:
        """
        Bind `callback` to a gadget property. When the property is updated, `callback`
        is called with no arguments.

        Parameters
        ----------
        prop : str
            The name of the gadget property.
        callback : Callable[[], None]
            Callback to bind to property.

        Returns
        -------
        int
            A unique id used to unbind the callback.
        """
        uid = next(_UID)
        setter = getattr(type(self), prop).fset
        bindings = setter.instances.setdefault(self, {})
        bindings[uid] = callback
        self.__bindings[uid] = prop
        return uid

    def unbind(self, uid: int) -> None:
        """
        Unbind a callback from a gadget property.

        Parameters
        ----------
        uid : int
            Unique id returned by the :meth:`bind` method.
        """
        prop = self.__bindings.pop(uid, None)
        if prop is None:
            return
        setter = getattr(type(self), prop).fset
        if self in setter.instances:
            setter.instances[self].pop(uid, None)

    @staticmethod
    def _tween_lerp(start, end, p) -> Real | Sequence[Real] | PosHint | SizeHint | None:
        """Lerp for none, sequences, and hints."""
        if start is None or end is None:
            return None

        if isinstance(start, Real) and isinstance(end, Real):
            value = lerp(start, end, p)
            if isinstance(start, int):
                return round_down(value)
            return value

        if isinstance(start, Sequence):
            return [
                Gadget._tween_lerp(start_value, end_value, p)
                for start_value, end_value in zip(start, end)
            ]

        if isinstance(start, MappingProxyType):
            return {
                key: Gadget._tween_lerp(start_value, end.get(key), p)
                for key, start_value in start.items()
            }

    async def tween(
        self,
        *,
        duration: float = 1.0,
        easing: Easing = "linear",
        on_start: Callable[[], None] | None = None,
        on_progress: Callable[[float], None] | None = None,
        on_complete: Callable[[], None] | None = None,
        **properties: Real | NDArray[np.number] | Sequence[Real] | PosHint | SizeHint,
    ) -> None:
        """
        Coroutine that sequentially updates gadget properties over a duration (in
        seconds).

        Parameters
        ----------
        duration : float, default: 1.0
            The duration of the tween in seconds.
        easing : Easing, default: "linear"
            The easing used for tweening.
        on_start : Callable[[], None] | None, default: None
            Called when tween starts.
        on_progress : Callable[[float], None] | None, default: None
            Called as tween updates with current progress.
        on_complete : Callable[[], None] | None, default: None
            Called when tween completes.
        **properties : Real | NDArray[np.number] | Sequence[Real] | PosHint | SizeHint
            Gadget properties' target values. E.g., to smoothly tween a gadget's
            position to (5, 10) over 2.5 seconds, specify the `pos` property as a
            keyword-argument:
            ``await gadget.tween(pos=(5, 10), duration=2.5, easing="out_bounce")``

        Notes
        -----
        Tweened values will be coerced to match the type of the initial value of their
        corresponding property.

        Non-numeric values will be set immediately.

        Warnings
        --------
        Running several tweens on the same properties concurrently will probably result
        in unexpected behavior.
        """
        end_time = perf_counter() + duration
        easing_function = EASINGS[easing]
        start_values = tuple(getattr(self, attr) for attr in properties)

        if pos_hint := properties.get("pos_hint"):
            properties["pos_hint"] = _normalize_pos_hint(pos_hint)
        if size_hint := properties.get("size_hint"):
            properties["size_hint"] = _normalize_size_hint(size_hint)
        if on_start is not None:
            on_start()

        while (current_time := perf_counter()) < end_time:
            p = easing_function(1 - (end_time - current_time) / duration)

            for start_value, (prop, target) in zip(start_values, properties.items()):
                setattr(self, prop, Gadget._tween_lerp(start_value, target, p))

            if on_progress is not None:
                on_progress(p)

            await asyncio.sleep(0)

        for prop, target in properties.items():
            setattr(self, prop, target)

        if on_complete is not None:
            on_complete()

    def on_size(self) -> None:
        """Update gadget after a resize."""

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""

    def on_add(self) -> None:
        """Update gadget after being added to the gadget tree."""
        self.apply_hints()
        for child in self.children:
            child.on_add()

    def on_remove(self) -> None:
        """Update gadget after being removed from the gadget tree."""
        for child in self.children:
            child.on_remove()

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """
        Handle a key press event.

        Handled key presses should return ``True``.

        Parameters
        ----------
        key_event : KeyEvent
            The key event to handle.

        Returns
        -------
        bool | None
            Whether the key event was handled.
        """

    def on_mouse(self, mouse_event: MouseEvent) -> bool | None:
        """
        Handle a mouse event.

        Handled mouse events should return ``True``.

        Parameters
        ----------
        mouse_event : MouseEvent
            The mouse event to handle.

        Returns
        -------
        bool | None
            Whether the mouse event was handled.
        """

    def on_paste(self, paste_event: PasteEvent) -> bool | None:
        """
        Handle a paste event.

        Handled paste events should return ``True``.

        Parameters
        ----------
        paste_event : PasteEvent
            The paste event to handle.

        Returns
        -------
        bool | None
            Whether the paste event was handled.
        """

    def on_terminal_focus(self, focus_event: FocusEvent) -> bool | None:
        """
        Handle a focus event.

        Handled focus events should return ``True``.

        Parameters
        ----------
        focus_event : FocusEvent
            The focus event to handle.

        Returns
        -------
        bool | None
            Whether the focus event was handled.
        """



"""A spotlight effect."""

import asyncio
from dataclasses import dataclass
from random import randrange

import numpy as np
from numpy.typing import NDArray

from ...colors import BLUE, GREEN, RED, Color
from ...geometry import BezierCurve, Point, Size, move_along_path
from ..text import Text


@dataclass
class _SpotLight:
    pos: Point
    color: Color
    radius: float = 8
    falloff: float = 0.3


async def spotlights_effect(text: Text):
    """
    Create a spotlight effect.

    Parameters
    ----------
    text : Text
        The text gadget which to run the effect.

    Warnings
    --------
    Modifying `text` size while effect is running will break the effect.
    """
    cover = Text(size=text.size)
    cover.canvas[:] = text.canvas

    in_fg = text.canvas["fg_color"]
    in_bg = text.canvas["bg_color"]
    out_fg = cover.canvas["fg_color"]
    out_bg = cover.canvas["bg_color"]

    spotlights = [
        _SpotLight(pos=_random_point(text.size), color=RED),
        _SpotLight(pos=_random_point(text.size), color=GREEN),
        _SpotLight(pos=_random_point(text.size), color=BLUE),
    ]

    text.add_gadget(cover)
    for _ in range(5):
        positions = [_random_point(text.size) for _ in spotlights]
        await _move_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights, positions)

    center = text.height // 2, text.width // 2
    positions = [center for _ in spotlights]
    await _move_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights, positions)
    await _grow_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights)
    text.remove_gadget(cover)


def _random_point(size: Size):
    h, w = size
    return randrange(h), randrange(w)


def _draw_spotlights(
    in_fg: NDArray[np.uint8],
    in_bg: NDArray[np.uint8],
    out_fg: NDArray[np.uint8],
    out_bg: NDArray[np.uint8],
    spotlights: list[_SpotLight],
):
    out_fg[:] = in_fg
    out_bg[:] = in_bg
    weighted = np.zeros_like(in_fg, float)
    for spotlight in spotlights:
        h, w, _ = in_fg.shape
        y, x = spotlight.pos
        x /= 2
        ys, xs = np.indices((h, w)).astype(float)
        xs /= 2

        distances = ((ys - y) ** 2 + (xs - x) ** 2) ** 0.5
        distances[distances > spotlight.radius] = np.inf
        weights = np.exp(-spotlight.falloff * distances)
        color = np.array(spotlight.color, float)
        color /= 255
        weighted += color * weights[:, :, None]

    weighted.clip(0, 1, out=weighted)
    out_fg[:] = (in_fg * weighted).astype(np.uint8)
    out_bg[:] = (in_bg * weighted).astype(np.uint8)


async def _move_spotlights(
    in_fg: NDArray[np.uint8],
    in_bg: NDArray[np.uint8],
    out_fg: NDArray[np.uint8],
    out_bg: NDArray[np.uint8],
    spotlights: list[_SpotLight],
    positions: list[Point],
):
    def draw(*args):
        _draw_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights)

    paths = []
    for spotlight, position in zip(spotlights, positions):
        curve = BezierCurve(np.array([spotlight.pos, position], float))
        paths.append([curve])

    motions = [
        move_along_path(
            spotlight,
            path,
            speed=20,
            easing="in_out_quad",
            on_start=draw,
            on_progress=draw,
            on_complete=draw,
        )
        for spotlight, path in zip(spotlights, paths)
    ]
    await asyncio.gather(*motions)


async def _grow_spotlights(
    in_fg: NDArray[np.uint8],
    in_bg: NDArray[np.uint8],
    out_fg: NDArray[np.uint8],
    out_bg: NDArray[np.uint8],
    spotlights: list[_SpotLight],
):
    h, w, _ = in_fg.shape

    end_radius = ((h / 2) ** 2 + (w / 2) ** 2) ** 0.5
    frames = 100

    all_radii = [
        np.linspace(spotlight.radius, end_radius, frames) for spotlight in spotlights
    ]
    all_falloffs = [
        np.linspace(spotlight.falloff, 0, frames) for spotlight in spotlights
    ]

    for i in range(frames):
        for spotlight, radii, falloffs in zip(spotlights, all_radii, all_falloffs):
            spotlight.radius = radii[i]
            spotlight.falloff = falloffs[i]

        _draw_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights)
        await asyncio.sleep(0.02)



"""A 2-D line plot gadget."""

from __future__ import annotations

from collections.abc import Sequence
from math import ceil
from numbers import Real

import cv2
import numpy as np
from uwcwidth import wcswidth

from ..colors import NEPTUNE_PRIMARY_BG, NEPTUNE_PRIMARY_FG, Color, rainbow_gradient
from ..terminal.events import MouseEvent
from .behaviors.movable import Movable
from .gadget import Gadget, Point, PosHint, Size, SizeHint, lerp
from .graphics import Blitter, Graphics, scale_geometry
from .pane import Pane
from .scroll_view import ScrollView
from .text import Text, add_text, new_cell

__all__ = ["LinePlot", "Point", "Size"]

PLOT_ZOOM = [1.0, 1.25, 1.5, 2.0, 3.0]
TICK_WIDTH = 11
VERTICAL_SPACING = 5
PRECISION = 4

# Derived
TICK_HALF = TICK_WIDTH // 2
VERTICAL_HALF = VERTICAL_SPACING // 2


class _Legend(Movable, Text):
    def _build_legend(self):
        plot: LinePlot = self.parent.parent
        colors = (
            rainbow_gradient(len(self.labels))
            if plot.line_colors is None
            else plot.line_colors
        )
        self.is_enabled = self.labels and len(self.labels) == len(colors)
        if self.is_enabled:
            height = len(self.labels) + 2
            width = max(map(wcswidth, self.labels)) + 6

            self.size = height, width
            self.canvas["fg_color"] = plot.plot_fg_color
            self.canvas["bg_color"] = plot.plot_bg_color
            self.chars[:] = " "

            self.add_border()

            self.chars[1:-1, 2] = "█"
            self.canvas["fg_color"][1:-1, 2] = colors

            for i, name in enumerate(self.labels, start=1):
                self.add_str(name, pos=(i, 4))


class _LinePlotProperty:
    def __set_name__(self, owner, name):
        self.name = f"_{name}"

    def __get__(self, instance, owner):
        if instance is None:
            return self

        return getattr(instance, self.name)

    def __set__(self, instance, value):
        if value == "sixel" and not Graphics._sixel_support:
            value = "half"
        setattr(instance, self.name, value)
        instance._build_plot()


class LinePlot(Gadget):
    r"""
    A 2-D line plot gadget.

    Zoom in or out with mouse wheel or "page_up"/"page_down".

    Parameters
    ----------
    xs : Sequence[Sequence[Real]]
        x-coordinates of each plot.
    ys : Sequence[Sequence[Real]]
        y-coordinates of each plot.
    blitter : Blitter, default: "braille"
        Determines how the line plot is rendered.
    min_x : Real | None, default: None
        Minimum x-value of plot. If `None`, min_x will be minimum of all xs.
    max_x : Real | None, default: None
        Maximum x-value of plot. If `None`, max_x will be maximum of all xs.
    min_y : Real | None, default: None
        Minimum y-value of plot. If `None`, min_y will be minimum of all ys.
    max_y : Real | None, default: None
        Maximum y-value of plot. If `None`, max_y will be maximum of all ys.
    line_colors : list[Color] | None, default: None
        The color of each line plot. If `None`, a rainbow gradient is used.
    legend_labels : list[str] | None, default: None
        Labels for legend. If `None`, legend is hidden.
    plot_fg_color : Color, default: NEPTUNE_PRIMARY_FG,
        Foreground color of the plot.
    plot_bg_color : Color, default: NEPTUNE_PRIMARY_BG,
        Background color of the plot.
    x_label : str | None, default: None
        Optional label for x-axis.
    y_label : str | None, default: None
        Optional label for y-axis.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    xs : Sequence[Sequence[Real]]
        x-coordinates of each plot.
    ys : Sequence[Sequence[Real]]
        y-coordinates of each plot.
    blitter : Blitter
        Determines how the line plot is rendered.
    min_x : Real | None
        Minimum x-value of plot. If `None`, min_x will be minimum of all xs.
    max_x : Real | None
        Maximum x-value of plot. If `None`, max_x will be maximum of all xs.
    min_y : Real | None
        Minimum y-value of plot. If `None`, min_y will be minimum of all ys.
    max_y : Real | None
        Maximum y-value of plot. If `None`, max_y will be maximum of all ys.
    line_colors : list[Color] | None
        The color of each line plot. If `None`, a rainbow gradient is used.
    legend_labels : list[str] | None
        Labels for legend. If `None`, legend is hidden.
    plot_fg_color : Color
        Foreground color of the plot.
    plot_bg_color : Color
        Background color of the plot.
    x_label : str | None
        Optional label for x-axis.
    y_label : str | None
        Optional label for y-axis.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    xs: Sequence[Sequence[Real]] = _LinePlotProperty()
    """x-coordinates of each plot."""
    ys: Sequence[Sequence[Real]] = _LinePlotProperty()
    """x-coordinates of each plot."""
    blitter: Blitter = _LinePlotProperty()
    """Determines which characters are used to draw the plot."""
    min_x: Real | None = _LinePlotProperty()
    """Minimum x-value of plot."""
    max_x: Real | None = _LinePlotProperty()
    """Maximum x-value of plot."""
    min_y: Real | None = _LinePlotProperty()
    """Minimum y-value of plot."""
    max_y: Real | None = _LinePlotProperty()
    """Maximum y-value of plot."""
    line_colors: list[Color] | None = _LinePlotProperty()
    """The color of each line plot."""

    def __init__(
        self,
        *,
        xs: Sequence[Sequence[Real]],
        ys: Sequence[Sequence[Real]],
        blitter: Blitter = "braille",
        min_x: Real | None = None,
        max_x: Real | None = None,
        min_y: Real | None = None,
        max_y: Real | None = None,
        line_colors: list[Color] | None = None,
        legend_labels: list[str] | None = None,
        plot_fg_color: Color = NEPTUNE_PRIMARY_FG,
        plot_bg_color: Color = NEPTUNE_PRIMARY_BG,
        x_label: str | None = None,
        y_label: str | None = None,
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        default_cell = new_cell(fg_color=plot_fg_color, bg_color=plot_bg_color)
        self._traces = Graphics(default_color=(*plot_bg_color, 0), is_transparent=True)
        self._scroll_view = ScrollView(
            show_vertical_bar=False,
            show_horizontal_bar=False,
            scrollwheel_enabled=False,
            alpha=0,
        )
        self._x_ticks = Text(default_cell=default_cell)
        self._y_ticks = Text(default_cell=default_cell)
        self._x_ticks_container = Gadget(is_transparent=True)
        self._y_ticks_container = Gadget(is_transparent=True)
        self._tick_corner = Text(size=(3, TICK_WIDTH + 1), default_cell=default_cell)
        self._x_label_gadget = Text(default_cell=default_cell)
        self._y_label_gadget = Text(default_cell=default_cell)
        self._legend = _Legend(allow_oob=False, is_enabled=False)
        self._container = Pane(
            size_hint={"height_hint": 1.0, "width_hint": 1.0}, bg_color=plot_bg_color
        )
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self._xs = xs
        self._ys = ys
        self._blitter = blitter
        self._min_x = min_x
        self._max_x = max_x
        self._min_y = min_y
        self._max_y = max_y
        self._line_colors = line_colors
        self._legend.labels = legend_labels
        self._plot_fg_color = plot_fg_color
        self._plot_bg_color = plot_bg_color
        self.alpha = alpha
        self.on_transparency()
        self.x_label = x_label
        self.y_label = y_label
        self._traces_zoom_index = 0
        """Index of size hint in `PLOT_ZOOM` that `_traces` is using."""

        def set_x_left():
            self._x_ticks.left = self._traces.left

        def set_y_top():
            self._y_ticks.top = self._traces.top

        self._traces.bind("pos", set_x_left)
        self._traces.bind("pos", set_y_top)
        self._tick_corner.chars[0, -1] = "└"

        self._scroll_view.view = self._traces
        self._x_ticks_container.add_gadget(self._x_ticks)
        self._y_ticks_container.add_gadget(self._y_ticks)
        self._container.add_gadgets(
            self._scroll_view,
            self._x_ticks_container,
            self._y_ticks_container,
            self._tick_corner,
            self._x_label_gadget,
            self._y_label_gadget,
            self._legend,
        )
        self.add_gadget(self._container)
        self._legend._build_legend()

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._scroll_view.is_transparent = self.is_transparent
        self._x_ticks.is_transparent = self.is_transparent
        self._y_ticks.is_transparent = self.is_transparent
        self._tick_corner.is_transparent = self.is_transparent
        self._x_label_gadget.is_transparent = self.is_transparent
        self._y_label_gadget.is_transparent = self.is_transparent
        self._container.is_transparent = self.is_transparent

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._container.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._container.alpha = alpha

    @property
    def plot_fg_color(self) -> Color:
        """Foreground color of text in the plot."""
        return self._plot_fg_color

    @plot_fg_color.setter
    def plot_fg_color(self, plot_fg_color: Color):
        self._plot_fg_color = plot_fg_color

        for child in self.walk():
            if isinstance(child, Text):
                child.canvas["fg_color"] = plot_fg_color
                child.default_fg_color = plot_fg_color
            elif isinstance(child, Pane):
                child.fg_color = plot_fg_color

        self._legend._build_legend()

    @property
    def plot_bg_color(self) -> Color:
        """Foreground color of text in the plot."""
        return self._plot_bg_color

    @plot_bg_color.setter
    def plot_bg_color(self, plot_bg_color: Color):
        self._plot_bg_color = plot_bg_color

        for child in self.walk():
            if isinstance(child, Text):
                child.canvas["bg_color"] = plot_bg_color
                child.default_bg_color = plot_bg_color
            elif isinstance(child, Pane):
                child.bg_color = plot_bg_color
            elif isinstance(child, Graphics):
                child.default_color = (*plot_bg_color, 0)

        self._legend._build_legend()

    @property
    def legend_labels(self) -> list[str] | None:
        """Labels for legend. If `None`, legend is hidden."""
        return self._legend.labels

    @legend_labels.setter
    def legend_labels(self, legend_labels: list[str] | None):
        self._legend.labels = legend_labels
        self._legend._build_legend()

    @property
    def x_label(self) -> str | None:
        """Optional label for x-axis."""
        return self._x_label

    @x_label.setter
    def x_label(self, x_label: str | None):
        self._x_label = x_label

        if x_label is not None:
            self._x_label_gadget.set_text(x_label)

        self._build_plot()

    @property
    def y_label(self) -> str | None:
        """Optional label for y-axis."""
        return self._y_label

    @y_label.setter
    def y_label(self, y_label: str | None):
        self._y_label = y_label

        if y_label is not None:
            self._y_label_gadget.size = wcswidth(y_label), 1
            add_text(self._y_label_gadget.canvas[:, 0], y_label)

        self._build_plot()

    def _build_plot(self):
        h, w = self.size
        if self.root is None or h == 0 or w == 0:
            return

        has_x_label = self._x_label_gadget.is_enabled = self.x_label is not None
        has_y_label = self._y_label_gadget.is_enabled = self.y_label is not None

        sv_left = has_y_label + TICK_WIDTH
        self._scroll_view.pos = 0, sv_left
        self._scroll_view.size = h - 2 - has_x_label, w - sv_left
        self._y_ticks_container.left = has_y_label
        self._y_ticks_container.size = self._scroll_view.height, TICK_WIDTH
        self._x_ticks_container.top = self._scroll_view.bottom
        self._x_ticks_container.left = self._y_ticks_container.right
        self._x_ticks_container.size = 2, self._scroll_view.width

        self._x_label_gadget.pos = (
            h - 1,
            self._scroll_view.width // 2 - self._x_label_gadget.width // 2 + TICK_WIDTH,
        )

        self._y_label_gadget.top = (
            self._scroll_view.height // 2 - self._y_label_gadget.height // 2
        )

        self._legend.top = self._scroll_view.height - self._legend.height
        self._legend.left = w - self._legend.width - ceil(TICK_WIDTH / 2)

        zoom = PLOT_ZOOM[self._traces_zoom_index]
        self._traces.size = (
            round(self._scroll_view.height * zoom),
            round(self._scroll_view.width * zoom),
        )

        offset_h = self._traces.height
        plot_right = self._traces.width - ceil(TICK_WIDTH / 2)
        offset_w = plot_right - TICK_HALF

        if offset_h <= 1 or offset_w <= 1:
            return

        self._traces.clear()
        if self._traces.blitter != self.blitter:
            self._traces.blitter = self.blitter

        min_x = min(xs.min() for xs in self.xs) if self.min_x is None else self.min_x
        max_x = max(xs.max() for xs in self.xs) if self.max_x is None else self.max_x
        min_y = min(ys.min() for ys in self.ys) if self.min_y is None else self.min_y
        max_y = max(ys.max() for ys in self.ys) if self.max_y is None else self.max_y

        if self.line_colors is None:
            line_colors = rainbow_gradient(len(self.xs))
        else:
            line_colors = self.line_colors

        x_delta = max_x - min_x
        y_delta = max_y - min_y

        plot_h, plot_w = scale_geometry(self._blitter, Size(offset_h, offset_w))
        plot = np.zeros((plot_h, plot_w, 4), np.uint8)
        for xs, ys, color in zip(self.xs, self.ys, line_colors, strict=True):
            scaled_ys = plot_h * (ys - min_y) / y_delta
            scaled_xs = plot_w * (xs - min_x) / x_delta
            coords = np.dstack((scaled_xs, plot_h - scaled_ys)).astype(int)
            cv2.polylines(plot, coords, isClosed=False, color=(*color, 255))
        _, left = scale_geometry(self._blitter, Size(1, TICK_HALF))
        _, right = scale_geometry(self._blitter, Size(1, plot_right))
        self._traces.texture[:, left:right] = plot

        # Regenerate Ticks
        self._y_ticks.size = self._traces.height, TICK_WIDTH
        self._y_ticks.canvas["fg_color"] = self.plot_fg_color
        self._y_ticks.canvas["bg_color"] = self.plot_bg_color
        self._y_ticks.chars[:, :-1] = " "
        self._y_ticks.chars[1:, -1] = "│"

        self._x_ticks.size = 2, self._traces.width
        self._x_ticks.canvas["fg_color"] = self.plot_fg_color
        self._x_ticks.canvas["bg_color"] = self.plot_bg_color
        self._x_ticks.chars[0, : plot_right - 1] = "─"
        self._x_ticks.chars[0, plot_right:] = " "
        self._x_ticks.chars[1:] = " "

        self._tick_corner.pos = h - 2 - has_x_label, sv_left - TICK_WIDTH - 1

        last_y = offset_h - 1
        for row in range(last_y, -1, -VERTICAL_SPACING):
            y_label = lerp(max_y, min_y, row / last_y)
            self._y_ticks.add_str(
                f"{y_label:>{TICK_WIDTH - 2}.{PRECISION}g} ┤"[:TICK_WIDTH],
                pos=(row, 0),
            )
        self._y_ticks.chars[0, -1] = "┐"

        last_x = offset_w - 1
        for column in range(0, offset_w, TICK_WIDTH):
            x_label = lerp(min_x, max_x, column / last_x)
            self._x_ticks.chars[0, column + TICK_HALF] = "┬"
            self._x_ticks.add_str(
                f"{x_label:^{TICK_WIDTH}.{PRECISION}g}"[:TICK_WIDTH],
                pos=(1, column),
            )
        self._x_ticks.chars[0, plot_right - 1] = "┐"

    def on_size(self):
        """Rebuild plot on resize."""
        self._build_plot()

    def _zoom_in(self):
        if self._traces_zoom_index < len(PLOT_ZOOM) - 1:
            self._traces_zoom_index += 1

    def _zoom_out(self):
        if self._traces_zoom_index > 0:
            self._traces_zoom_index -= 1

    def on_mouse(self, mouse_event: MouseEvent) -> bool | None:
        """Zoom-in or -out on mouse wheel."""
        if not self.collides_point(mouse_event.pos):
            return

        if mouse_event.event_type == "scroll_up":
            self._zoom_in()
        elif mouse_event.event_type == "scroll_down":
            self._zoom_out()
        else:
            return super().on_mouse(mouse_event)

        self._build_plot()
        return True

    def on_key(self, key_event):
        """Zoom-in or -out with "page_up" or "page_down"."""
        if key_event.key == "page_up":
            self._zoom_in()
        elif key_event.key == "page_down":
            self._zoom_out()
        else:
            return super().on_key(key_event)

        self._build_plot()
        return True

    def on_add(self):
        """Built plot on add."""
        super().on_add()
        self._build_plot()



"""A grid layout gadget."""

from itertools import accumulate, product
from typing import Literal

from .gadget import Gadget, Point, PosHint, Size, SizeHint

__all__ = ["GridLayout", "Orientation", "Point", "Size"]

Orientation = Literal[
    "lr-tb", "lr-bt", "rl-tb", "rl-bt", "tb-lr", "tb-rl", "bt-lr", "bt-rl"
]
"""
Orientation of the grid.

Describes how the grid fills as children are added. As an example, the orientation
"lr-tb" means left-to-right, then top-to-bottom.
"""


class _RepositionProperty:
    def __set_name__(self, owner, name):
        self.name = "_" + name

    def __get__(self, instance, owner):
        if instance is None:
            return self

        return getattr(instance, self.name)

    def __set__(self, instance, value):
        setattr(instance, self.name, value)
        instance._reposition_children()


class GridLayout(Gadget):
    r"""
    A gadget that automatically positions children into a grid.

    Parameters
    ----------
    grid_rows : int, default: 1
        Number of rows.
    grid_columns : int, default: 1
        Number of columns.
    orientation : Orientation, default: "lr-tb"
        The orientation of the grid.
    padding_left : int, default: 0
        Padding on left side of grid.
    padding_right : int, default: 0
        Padding on right side of grid.
    padding_top : int, default: 0
        Padding at the top of grid.
    padding_bottom : int, default: 0
        Padding at the bottom of grid.
    horizontal_spacing : int, default: 0
        Horizontal spacing between children.
    vertical_spacing : int, default: 0
        Vertical spacing between children.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    min_grid_size : Size
        Minimum grid size needed to show all children.
    grid_rows : int
        Number of rows.
    grid_columns : int
        Number of columns.
    orientation : Orientation
        The orientation of the grid.
    padding_left : int
        Padding on left side of grid.
    padding_right : int
        Padding on right side of grid.
    padding_top : int
        Padding at the top of grid.
    padding_bottom : int
        Padding at the bottom of grid.
    horizontal_spacing : int
        Horizontal spacing between children.
    vertical_spacing : int
        Vertical spacing between children.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    index_at(row, col)
        Return index of the child gadget in :attr:`children` at position `row, col` in
        the grid.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.

    Notes
    -----
    Grid layouts remove size and pos hints from their children.

    Re-ordering children (such as through :meth:`pull_to_front`) and calling
    :meth:`_reposition_children` will change the positions of the children in the grid.

    The read-only attribute :attr:`min_grid_size` is the minimum size the grid must
    be to show all children. This can be used to set the size of the grid layout, e.g.,
    ``my_grid.size = my_grid.min_grid_size``.
    """

    grid_rows: int = _RepositionProperty()
    grid_columns: int = _RepositionProperty()
    padding_left: int = _RepositionProperty()
    padding_right: int = _RepositionProperty()
    padding_top: int = _RepositionProperty()
    padding_bottom: int = _RepositionProperty()
    horizontal_spacing: int = _RepositionProperty()
    vertical_spacing: int = _RepositionProperty()

    def __init__(
        self,
        *,
        grid_rows: int = 1,
        grid_columns: int = 1,
        orientation: Orientation = "lr-tb",
        padding_left: int = 0,
        padding_right: int = 0,
        padding_top: int = 0,
        padding_bottom: int = 0,
        horizontal_spacing: int = 0,
        vertical_spacing: int = 0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._grid_rows = grid_rows
        self._grid_columns = grid_columns
        self._orientation = orientation
        self._padding_left = padding_left
        self._padding_right = padding_right
        self._padding_top = padding_top
        self._padding_bottom = padding_bottom
        self._horizontal_spacing = horizontal_spacing
        self._vertical_spacing = vertical_spacing
        self._min_grid_size = Size(0, 0)

        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

    @property
    def orientation(self) -> Orientation:
        """
        The orientation of the grid.

        Describes how the grid fills as children are added. The default is left-to-right
        then top-to-bottom.
        """
        return self._orientation

    @orientation.setter
    def orientation(self, orientation: Orientation):
        if self._orientation not in Orientation.__args__:
            raise TypeError(f"{orientation} is not a valid orientation.")
        self._orientation = orientation
        self._reposition_children()

    def on_size(self):
        """Repositon children on resize."""
        self._reposition_children()

    def index_at(self, row: int, col: int) -> int:
        """
        Return the index of the child gadget in :attr:`children` at a given row and
        column in the grid.

        Parameters
        ----------
        row : int
            The row of the child.
        col : int
            The column of the child.

        Returns
        -------
        int
            Index of the child in :attr:`children`.
        """
        rows = self.grid_rows
        cols = self.grid_columns

        if self.orientation == "lr-tb":
            return col + row * cols
        if self.orientation == "lr-bt":
            return col + (rows - row - 1) * cols
        if self.orientation == "rl-tb":
            return (cols - col - 1) + row * cols
        if self.orientation == "rl-bt":
            return (cols - col - 1) + (rows - row - 1) * cols
        if self.orientation == "tb-lr":
            return row + col * rows
        if self.orientation == "tb-rl":
            return row + (cols - col - 1) * rows
        if self.orientation == "bt-lr":
            return (rows - row - 1) + col * rows
        if self.orientation == "bt-rl":
            return (rows - row - 1) + (cols - col - 1) * rows

    def _row_height(self, i: int) -> int:
        """Height of row `i`."""
        return max(
            (
                self.children[index].height
                for col in range(self.grid_columns)
                if (index := self.index_at(i, col)) < len(self.children)
            ),
            default=0,
        )

    def _col_width(self, i: int) -> int:
        """Width of column `i`."""
        return max(
            (
                self.children[index].width
                for row in range(self.grid_rows)
                if (index := self.index_at(row, i)) < len(self.children)
            ),
            default=0,
        )

    @property
    def min_grid_size(self) -> Size:
        """Return the minimum grid size to show all children."""
        nrows, ncols = self.grid_rows, self.grid_columns
        if nrows == 0 or ncols == 0:
            return Size(0, 0)

        bottom = (
            self.padding_top
            + sum(self._row_height(i) for i in range(nrows))
            + self.vertical_spacing * (nrows - 1)
            + self.padding_bottom
        )
        right = (
            self.padding_left
            + sum(self._col_width(i) for i in range(ncols))
            + self.horizontal_spacing * (ncols - 1)
            + self.padding_right
        )

        return Size(bottom, right)

    def _reposition_children(self):
        if self.grid_rows == 0 or self.grid_columns == 0:
            return

        row_tops = tuple(
            accumulate(
                self.padding_top
                if i == 0
                else self._row_height(i - 1) + self.vertical_spacing
                for i in range(self.grid_rows)
            )
        )
        col_lefts = tuple(
            accumulate(
                self.padding_left
                if i == 0
                else self._col_width(i - 1) + self.horizontal_spacing
                for i in range(self.grid_columns)
            )
        )

        for row, col in product(range(self.grid_rows), range(self.grid_columns)):
            if (i := self.index_at(row, col)) < len(self.children):
                self.children[i].pos = row_tops[row], col_lefts[col]

    def add_gadget(self, gadget):
        """Check to see if grid is full before adding gadget and remove its hints."""
        if len(self.children) >= self.grid_rows * self.grid_columns:
            raise ValueError("too many children, grid is full")

        gadget.size_hint = {}
        gadget.pos_hint = {}
        super().add_gadget(gadget)

        self._reposition_children()

    def remove_gadget(self, gadget):
        """Reposition children when a gadget is removed."""
        super().remove_gadget(gadget)
        self._reposition_children()
